import sys

from .core.cache import cache
from .lib import console, __pkg__, __version__
from .tui.menus import Menus


def start():
    try:
        console.set_window_title(title=f"{__pkg__.title()} v{__version__}")
        menu = Menus()
        menu.main()
    except KeyboardInterrupt:
        sys.exit()
    finally:
        cache.clear()



import csv
import json
import os
import subprocess
import typing as t
from datetime import datetime
from pathlib import Path

import pyfiglet
from prompt_toolkit.shortcuts import message_dialog
from rich.console import Console
from rich.text import Text
from rich.tree import Tree
from update_checker import UpdateChecker

from . import __pkg__, __version__

__all__ = [
    "__pkg__",
    "__version__",
    "console",
    "preview_response",
    "export_response",
    "check_updates",
    "clear_screen",
    "ascii_banner",
    "set_menu_title",
]

console = Console(log_time=False)


def preview_response(data: t.Union[dict, list], source: str, _type: str):
    """
    Display a preview of response data in a tree structure.

    :param data: The data to preview (dict or list).
    :param source: The source identifier of the data.
    :param _type: The type of data being previewed.
    """

    if isinstance(data, dict):
        tree = Tree(
            label=f"\n[bold]{data.get('name') or data.get('login') or data.get('id') or 'Data'}[/bold]",
            guide_style="#444444",
            highlight=True,
        )
        fill_tree(tree=tree, data=data)
        console.print(tree)
        print()

    elif isinstance(data, list):
        preview_data = data[:5]
        tree = Tree(
            label=f"\n[bold]First {len(preview_data)} of {len(data)} {_type} for '{source}'[/bold]",
            guide_style="#444444",
            highlight=True,
        )

        for item in preview_data:
            if isinstance(item, dict):
                branch_label = (
                    item.get("full_name")
                    or item.get("name")
                    or item.get("login")
                    or item.get("type")
                    or item.get("id")
                    or "Item"
                )
                branch = tree.add(label=f"[bold]{branch_label}[/bold]", highlight=True)
                fill_tree(tree=branch, data=item)

        console.print(tree)
        print()
    else:
        console.print(data)


def export_response(
    data: t.Union[dict, list],
    data_type: str,
    source: str,
    file_formats: list,
    output_dir: str = "../exports",
):
    """
    Export response data to one or more file formats.

    :param data: The data to export (dict or list).
    :param data_type: The type of data being exported.
    :param source: The source identifier of the data.
    :param file_formats: List of file formats to export to (json, csv, html).
    :param output_dir: Directory path where exported files will be saved.
    :return: List of exported file paths.
    """

    # Create output directory if it doesn't exist
    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    # Generate filename with timestamp
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    safe_source = source.replace("/", "_")
    filename = f"{safe_source}_{data_type}_{timestamp}"

    # Normalize data to list of dicts
    if isinstance(data, dict):
        data_list = [data]
    else:
        data_list = data

    # Export to all selected formats
    exported_files = []
    for file_format in file_formats:
        filepath = output_dir / f"{filename}.{file_format}"

        if file_format == "json":
            with open(filepath, "w", encoding="utf-8") as f:
                json.dump(data_list, f, indent=2, ensure_ascii=False)

        elif file_format == "csv":
            if data_list:
                # Get all unique keys from all dicts
                keys = set()
                for item in data_list:
                    if isinstance(item, dict):
                        keys.update(item.keys())
                keys = sorted(keys)

                with open(filepath, "w", newline="", encoding="utf-8") as f:
                    writer = csv.DictWriter(f, fieldnames=keys)
                    writer.writeheader()
                    for item in data_list:
                        if isinstance(item, dict):
                            writer.writerow(item)

        elif file_format == "html":
            with open(filepath, "w", encoding="utf-8") as file:
                file.write("<!DOCTYPE html>\n<html>\n<head>\n")
                file.write('<meta charset="UTF-8">\n')
                file.write("<style>table { border-collapse: collapse; width: 100%; }\n")
                file.write(
                    "th, td { border: 1px solid black; padding: 8px; text-align: left; }\n"
                )
                file.write("th { background-color: #f2f2f2; }</style>\n")
                file.write("</head>\n<body>\n")
                file.write('<table class="table table-striped" border="1">\n')

                if data_list:
                    # Get all unique keys
                    keys = set()
                    for item in data_list:
                        if isinstance(item, dict):
                            keys.update(item.keys())
                    keys = sorted(keys)

                    # Write header
                    file.write("<thead>\n<tr>\n")
                    for key in keys:
                        file.write(f"<th>{key}</th>\n")
                    file.write("</tr>\n</thead>\n")

                    # Write rows
                    file.write("<tbody>\n")
                    for item in data_list:
                        if isinstance(item, dict):
                            file.write("<tr>\n")
                            for key in keys:
                                value = item.get(key, "")
                                file.write(f"<td>{value}</td>\n")
                            file.write("</tr>\n")
                    file.write("</tbody>\n")

                file.write("</table>\n</body>\n</html>")

        exported_files.append(str(filepath))

    # Show success message
    console.print("\n[green]✓ Export successful![/green]")
    for filepath in exported_files:
        console.print(f"  • {filepath}")

    return exported_files


def fill_tree(tree: Tree, data: t.Union[dict, list]) -> Tree:
    """
    Recursively populate a Rich Tree with data.

    :param tree: The Tree object to populate.
    :param data: The data to add to the tree (dict or list).
    :return: The populated Tree object.
    """

    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, dict) or isinstance(value, list):
                branch = tree.add(f"{key}")
                fill_tree(tree=branch, data=value)
            else:
                tree.add(f"[dim]{key.replace('_', ' ').title()}[/dim]: {value}")

    elif isinstance(data, list):
        for item in data:
            if isinstance(item, dict) or isinstance(item, list):
                fill_tree(tree=tree, data=item)
            else:
                tree.add(str(item))

    else:
        tree.add(str(data))

    return tree


def check_updates():
    """Check for available package updates and display the result."""

    with console.status("[dim]Checking for updates[/dim]…") as status:
        checker = UpdateChecker()
        result = checker.check(__pkg__, __version__)
        if result is not None:
            status.stop()
            message_dialog(title="Update Available", text=result).run()
        else:
            status.stop()
            message_dialog(
                title="Up to Date",
                text=f"You're running the current version, {__version__}",
            ).run()


def clear_screen():
    """Clear the terminal screen."""
    subprocess.run(["cls" if os.name == "nt" else "clear"])


def ascii_banner(text: str):
    """Display a colourful ASCII art banner with gradient styling.

    :param text: The text to convert to ASCII art.
    """
    clear_screen()

    ascii_text = pyfiglet.figlet_format(text=text, font="chunky")
    banner_text = Text(ascii_text)

    length = len(ascii_text)
    for i in range(length):
        ratio = i / max(length - 1, 1)
        r = int(255 - 127 * ratio)
        g = int(200 - 200 * ratio)
        b = 255

        banner_text.stylize(f"rgb({r},{g},{b})", i, i + 1)
    console.print(banner_text)


def set_menu_title(menu_type: t.Literal["home", "user", "org", "repo", "search"]):
    """
    Set the terminal window title based on the current menu.

    :param menu_type: The type of menu being displayed.
    """

    title: str = __pkg__.title()
    title += f" | {menu_type.title()}"
    console.set_window_title(title=title)



import sys
import typing as t

import questionary as q
from questionary import Style
from rich.status import Status

from .dialogs import Dialogs
from .prompts import Prompts
from ..core.cache import cache
from ..core.models import User, Repo, Org, Search
from ..lib import check_updates, preview_response, export_response, set_menu_title
from ..lib import console, clear_screen, ascii_banner

CUSTOM_STYLE = Style(
    [
        ("highlighted", "fg:black bold bg:white"),
        ("instruction", "fg:gray"),
    ]
)

INSTRUCTIONS = "↑↓ [move] • ⮠ [select]"
EXPORT_INSTRUCTIONS = "↑↓ [move] • ⮠ [confirm] • spacebar [check]"

dialogs = Dialogs()
prompts = Prompts()

__all__ = ["Menus"]


class BaseMenu:
    """Base class providing common menu functionality and response handling for GitHub data queries."""

    def __init__(self, main_menu: t.Callable):
        """
        Initialise the BaseMenu with pagination and search method configurations.

        :param main_menu: Callable reference to the main menu function.
        """

        # Define which methods require pagination
        self.paginated_methods = {
            "repos",
            "subscriptions",
            "starred",
            "followers",
            "following",
            "orgs",
            "gists",
            "events",
            "received_events",
            "forks",
            "issue_events",
            "assignees",
            "branches",
            "tags",
            "stargazers",
            "subscribers",
            "commits",
            "comments",
            "issues",
            "releases",
            "deployments",
            "labels",
            "hooks",
            "members",
        }

        # Methods that don't need pagination
        self.non_paginated_methods = {"profile", "languages"}

        # Search methods (all require pagination)
        self.search_methods = {"repos", "users", "commits", "issues", "topics"}

        self.main_menu = main_menu

    @staticmethod
    def target_validator(
        identifier: str,
        instance: t.Union[User, Repo, Org],
        callback: t.Callable,
        *callback_args,
    ) -> bool:
        """
        Validate whether a GitHub entity exists.

        :param identifier: Name or identifier of the entity to validate.
        :param instance: The instance with an exists() method (User, Repo, or Org).
        :param callback: Function to call if validation fails.
        :param callback_args: Arguments to pass to the callback function.
        :return: True if the entity exists, False otherwise.
        """

        if isinstance(instance, User):
            target_type = "user"
        elif isinstance(instance, Repo):
            target_type = "repo"
        else:
            target_type = "org"

        with Status(
            f"[dim]Validating {target_type} ({identifier})[/dim]…",
            console=console,
        ) as status:
            exists, response = instance.exists()

            if not exists:
                status.stop()
                if response and "message" in response:
                    console.print(
                        f"[bold][yellow]✘[/yellow] {response['message']}[/bold]"
                    )
                console.input("  Press [bold]ENTER[/bold] to continue …")
                callback(*callback_args)
                return False

            return True

    def execute_and_handle_response(
        self,
        instance: t.Union[User, Repo, Org, Search],
        method_name: str,
        target_type: t.Literal["user", "repo", "org"],
        source: str,
    ):
        """
        Execute a method on an instance and handle the resulting data.

        :param instance: The instance to execute the method on.
        :param method_name: Name of the method to execute.
        :param target_type: Type of entity ("user", "repo", or "org").
        :param source: Source identifier for response handling.
        """

        valid_methods = self.paginated_methods | self.non_paginated_methods
        if method_name in valid_methods:
            with Status(
                status=f"[dim]Initialising {target_type} {method_name}[/dim]…",
                console=console,
            ) as status:
                data = self.execute_selection(
                    source=source,
                    instance=instance,
                    method_name=method_name,
                    status=status,
                )

                status.stop()
                self.response_handler(data=data, data_type=method_name, source=source)

    def execute_selection(self, status: Status, **kwargs):
        """
        Execute a method on an instance with pagination if required.

        :param status: Rich Status object for displaying progress.
        :param kwargs: Must include 'instance', 'method_name', and 'source'.
        :return: The data returned by the executed method.
        """

        instance = kwargs.get("instance")
        method_name = kwargs.get("method_name")
        source = kwargs.get("source")

        method = getattr(instance, method_name)
        status.stop()
        # pagination params if needed
        params = (
            prompts.pagination_params() if method_name in self.paginated_methods else {}
        )
        status.start()
        status.update(f"[dim]Getting {method_name} from {source}[/dim]…")
        return method(**params)

    def response_handler(self, data: t.Union[dict, list], data_type: str, source: str):
        """
        Handle retrieved data by previewing and offering export options.

        :param data: The data to handle (dict or list).
        :param data_type: Type of data retrieved.
        :param source: Source identifier of the data.
        """

        try:
            if not data:
                console.print(
                    f"[bold][yellow]✘[/yellow] No data found for '{source}'[/bold]"
                )
            else:
                preview_response(data=data, source=source, _type=data_type)
                export_choice = q.select(
                    "What would you like to do?",
                    choices=[
                        q.Choice(
                            title="Export",
                            value="export",
                            description="Export the data",
                            shortcut_key="e",
                        ),
                        q.Choice(
                            title="Skip",
                            value="skip",
                            description="Do nothing, and go back to previous menu",
                            shortcut_key="x",
                        ),
                        q.Choice(
                            title="Quit",
                            value="quit",
                            description="Close this session",
                        ),
                    ],
                    style=CUSTOM_STYLE,
                    instruction=INSTRUCTIONS,
                ).ask()

                if export_choice == "skip":
                    return

                if export_choice == "quit":
                    if dialogs.quit():
                        sys.exit()
                    else:
                        clear_screen()
                        self.response_handler(
                            data=data,
                            data_type=data_type,
                            source=source,
                        )
                        return

                file_formats = q.checkbox(
                    "Select export format(s)",
                    choices=[
                        q.Choice(
                            title="JSON",
                            value="json",
                            description="Export as JSON file",
                        ),
                        q.Choice(
                            title="CSV",
                            value="csv",
                            description="Export as CSV file",
                        ),
                        q.Choice(
                            title="HTML",
                            value="html",
                            description="Export as HTML table",
                        ),
                    ],
                    style=CUSTOM_STYLE,
                    instruction=EXPORT_INSTRUCTIONS,
                    validate=lambda x: len(x) > 0
                    or "Please select at least one format",
                ).ask()

                export_response(
                    data=data,
                    data_type=data_type,
                    source=source,
                    file_formats=file_formats,
                )

            console.input("  Press [bold]ENTER[/bold] to continue …")
        except KeyboardInterrupt:
            console.print("\nExport cancelled")

    def navigation_handler(self, option: str, callback: t.Callable, *callback_args):
        """
        Handle common navigation options across menus.

        :param option: The selected navigation option.
        :param callback: Function to call for certain navigation actions.
        :param callback_args: Arguments to pass to the callback function.
        :return: True if navigation was handled, False otherwise.
        """

        navigation_handlers = {
            "back": lambda: self.main_menu(),
            "quit": lambda: (
                sys.exit() if dialogs.quit() else callback(*callback_args)
            ),
        }

        handler = navigation_handlers.get(option)
        if handler:
            handler()
            return True
        return False


class Menus(BaseMenu):
    """Main menu system providing interactive interfaces for GitHub data queries."""

    def __init__(self):
        """Initialise the Menus class with mode handlers for different query types."""

        super().__init__(main_menu=self.main)

        self.mode_handlers = {
            "user": self.user,
            "repo": self.repo,
            "org": self.org,
            "search": self.search,
        }

    def main(self):
        """Display the main menu for selecting query mode."""

        set_menu_title(menu_type="home")
        clear_screen()
        try:
            ascii_banner(text="octosuite")

            action = q.select(
                "What action would you like to perform?",
                choices=[
                    q.Choice(
                        title="User",
                        value="user",
                        description="data about a user",
                    ),
                    q.Choice(
                        title="Repo",
                        value="repo",
                        description="data about a repository",
                    ),
                    q.Choice(
                        title="Org",
                        value="org",
                        description="data about an organisation",
                    ),
                    q.Choice(
                        title="Search",
                        value="search",
                        description="Search GitHub",
                    ),
                    q.Choice(
                        title="Clear cache",
                        value="clear_cache",
                        description="Clear all in-memory cache from Octosuite",
                    ),
                    q.Choice(
                        title="Updates",
                        value="updates",
                        description="Check for updates",
                    ),
                    q.Choice(
                        title="License",
                        value="license",
                        description="Read license notice, and copyright information",
                    ),
                    q.Choice(
                        title="Quit",
                        value="quit",
                        description="Close this session",
                    ),
                ],
                style=CUSTOM_STYLE,
                instruction=INSTRUCTIONS,
            ).ask()

            if action == "quit":
                if dialogs.quit():
                    sys.exit()
                else:
                    self.main()
            elif action == "license":
                dialogs.license()
                self.main()
            elif action == "updates":
                check_updates()
                self.main()
            elif action == "clear_cache":
                if dialogs.clear_cache():
                    cache.clear()
                self.main()
            elif action is None:
                sys.exit()
            else:
                handler = self.mode_handlers.get(action)
                if handler:
                    handler()
                else:
                    self.main()
        except KeyboardInterrupt:
            sys.exit()

    def search(self, query: t.Optional[str] = None):
        """
        Display the search menu for querying GitHub content.

        :param query: Optional search query string. If None, prompts user for input.
        """

        set_menu_title(menu_type="search")
        clear_screen()
        if query is None:
            ascii_banner(text="Search")
            query = prompts.prompt(
                message="Search Query",
                instruction="e.g., machine learning",
                style=CUSTOM_STYLE,
            )

        ascii_banner(text=query)
        option = q.select(
            "What would you like to do/search?",
            choices=[
                q.Choice(
                    title="Change Query",
                    value="change_query",
                    description="Search with a different query",
                    shortcut_key="c",
                ),
                q.Choice(
                    title="Repositories",
                    value="repos",
                    description="Search for repositories",
                ),
                q.Choice(
                    title="Users",
                    value="users",
                    description="Search for users",
                ),
                q.Choice(
                    title="Commits",
                    value="commits",
                    description="Search for commits",
                ),
                q.Choice(
                    title="Issues",
                    value="issues",
                    description="Search for issues and pull requests",
                ),
                q.Choice(
                    title="Topics",
                    value="topics",
                    description="Search for topics",
                ),
                q.Choice(
                    title="Go Back",
                    value="back",
                    description="Go back to previous menu",
                    shortcut_key="b",
                ),
                q.Choice(
                    title="Quit",
                    value="quit",
                    description="Close this session",
                    shortcut_key="q",
                ),
            ],
            style=CUSTOM_STYLE,
            instruction=INSTRUCTIONS,
            use_shortcuts=True,
        ).ask()

        if option is None:
            self.main()
            return

        # Handle navigation
        if self.navigation_handler(option, self.search, query):
            return

        # Handle change query
        if option == "change_query":
            self.search()
            return

        # Execute search if it's a valid method
        if option in self.search_methods:
            with Status(
                status=f"[dim]Initialising {option} search[/dim]…", console=console
            ) as status:
                status.stop()
                params = prompts.pagination_params()
                status.start()

                # Create Search instance
                search = Search(
                    query=query,
                    page=params["page"],
                    per_page=params["per_page"],
                )

                method = getattr(search, option)
                status.update(f"[dim]Searching {option} for {query}[/dim]…")
                data = method()

                if data:
                    items = data.get("items")
                    status.stop()
                    self.response_handler(
                        data=items if items is not None else data,
                        data_type=option,
                        source=query,
                    )

        # After handling response, show menu again
        self.search(query=query)

    def user(self, username: t.Optional[str] = None, is_validated: bool = False):
        """
        Display the user menu for querying GitHub user data.

        :param username: Optional GitHub username. If None, prompts user for input.
        :param is_validated: Whether the username has already been validated.
        """

        set_menu_title(menu_type="user")
        clear_screen()
        if username is None:
            ascii_banner(text="User")
            username = prompts.prompt(
                message="GitHub Username",
                instruction="e.g., octocat",
                style=CUSTOM_STYLE,
                qmark="@",
            )

        user = User(name=username)

        if not self.target_validator(
            identifier=username,
            instance=user,
            callback=self.user,
        ):
            return

        ascii_banner(text=username)
        option = q.select(
            "What would you like to do/get?",
            choices=[
                q.Choice(
                    title="Change Username",
                    value="change_username",
                    description="Query a different user",
                    shortcut_key="u",
                ),
                q.Choice(
                    title="Profile",
                    value="profile",
                    description="Profile data",
                ),
                q.Choice(
                    title="Repositories",
                    value="repos",
                    description="Public repositories",
                ),
                q.Choice(
                    title="Subscriptions",
                    value="subscriptions",
                    description="Subscribed repositories",
                ),
                q.Choice(
                    title="Starred",
                    value="starred",
                    description="Starred repositories",
                ),
                q.Choice(
                    title="Followers",
                    value="followers",
                    description="Accounts that follow this user",
                ),
                q.Choice(
                    title="Following",
                    value="following",
                    description="Accounts that this user follows",
                ),
                q.Choice(
                    title="Organisations",
                    value="orgs",
                    description="This user's organisations",
                ),
                q.Choice(
                    title="Gists",
                    value="gists",
                    description="Gists (code snippets/files)",
                ),
                q.Choice(
                    title="Events",
                    value="events",
                    description="Events",
                ),
                q.Choice(
                    title="Received Events",
                    value="received_events",
                    description="This user's received events",
                ),
                q.Choice(
                    title="Go Back",
                    value="back",
                    description="Go back to previous menu",
                    shortcut_key="b",
                ),
                q.Choice(
                    title="Quit",
                    value="quit",
                    description="Close this session",
                    shortcut_key="q",
                ),
            ],
            style=CUSTOM_STYLE,
            instruction=INSTRUCTIONS,
            use_shortcuts=True,
        ).ask()

        if option is None:
            self.main()
            return

        # Handle navigation
        elif self.navigation_handler(option, self.user, username):
            return

        # Handle change username
        elif option == "change_username":
            self.user()
            return
        else:
            # Execute action and handle response
            self.execute_and_handle_response(
                instance=user,
                method_name=option,
                target_type="user",
                source=username,
            )

            # After handling response, show menu again WITHOUT re-validating
            self.user(username=username, is_validated=True)

    def repo(
        self,
        name: t.Optional[str] = None,
        owner: t.Optional[str] = None,
    ):
        """
        Display the repository menu for querying GitHub repository data.

        :param name: Optional repository name. If None, prompts user for input.
        :param owner: Optional repository owner. If None, prompts user for input.
        """

        set_menu_title(menu_type="repo")
        clear_screen()
        if name is None or owner is None:
            ascii_banner(text="Repo")
            if name is None:
                name = prompts.prompt(
                    message="GitHub Repo Name",
                    instruction="e.g., spoon-knife",
                    style=CUSTOM_STYLE,
                )
            if owner is None:
                owner = prompts.prompt(
                    message="GitHub Repo Owner",
                    instruction="e.g., octocat",
                    style=CUSTOM_STYLE,
                    qmark="@",
                )

        repo = Repo(name=name, owner=owner)
        source = f"{owner}/{name}"

        # Only validate if not already validated
        if not self.target_validator(
            identifier=source,
            instance=repo,
            callback=self.repo,
        ):
            return

        ascii_banner(text=source)
        option = q.select(
            "What would you like to do/get?",
            choices=[
                q.Choice(
                    title="Change Repo Name",
                    value="change_repo_name",
                    description="Update the name of the repo",
                ),
                q.Choice(
                    title="Change Owner Name",
                    value="change_owner",
                    description="Update the name of repo owner",
                ),
                q.Choice(
                    title="Change Repo and Owner Names",
                    value="change_both",
                    description="Update both the repo and owner names",
                ),
                q.Choice(
                    title="Profile",
                    value="profile",
                    description="Repository data",
                ),
                q.Choice(
                    title="Forks",
                    value="forks",
                    description="Repository forks",
                ),
                q.Choice(
                    title="Issue Events",
                    value="issue_events",
                    description="Issue events",
                ),
                q.Choice(
                    title="Events",
                    value="events",
                    description="Repository events",
                ),
                q.Choice(
                    title="Assignees",
                    value="assignees",
                    description="Assignees",
                ),
                q.Choice(
                    title="Branches",
                    value="branches",
                    description="Branches",
                ),
                q.Choice(
                    title="Tags",
                    value="tags",
                    description="Tags",
                ),
                q.Choice(
                    title="Languages",
                    value="languages",
                    description="Programming languages",
                ),
                q.Choice(
                    title="Stargazers",
                    value="stargazers",
                    description="Users who starred this repo",
                ),
                q.Choice(
                    title="Subscribers",
                    value="subscribers",
                    description="Users subscribed to this repo",
                ),
                q.Choice(
                    title="Commits",
                    value="commits",
                    description="Commits",
                ),
                q.Choice(
                    title="Comments",
                    value="comments",
                    description="Comments",
                ),
                q.Choice(
                    title="Issues",
                    value="issues",
                    description="Issues",
                ),
                q.Choice(
                    title="Releases",
                    value="releases",
                    description="Releases",
                ),
                q.Choice(
                    title="Deployments",
                    value="deployments",
                    description="Deployments",
                ),
                q.Choice(
                    title="Labels",
                    value="labels",
                    description="Labels",
                ),
                q.Choice(
                    title="Go Back",
                    value="back",
                    description="Go back to previous menu",
                    shortcut_key="b",
                ),
                q.Choice(
                    title="Quit",
                    value="quit",
                    description="Close this session",
                    shortcut_key="q",
                ),
            ],
            style=CUSTOM_STYLE,
            instruction=INSTRUCTIONS,
            use_shortcuts=True,
        ).ask()

        # Handle change options
        change_handlers = {
            "change_repo_name": lambda: self.repo(owner=owner),
            "change_owner": lambda: self.repo(name=name),
            "change_both": lambda: self.repo(),
        }

        if option is None:
            self.main()
            return

        # Handle navigation
        elif self.navigation_handler(option, self.repo, name, owner):
            return

        elif option in change_handlers:
            change_handlers[option]()
            return
        else:
            # Execute action and handle response
            self.execute_and_handle_response(
                instance=repo,
                method_name=option,
                target_type="repo",
                source=source,
            )

            # After handling response, show menu again WITHOUT re-validating
            self.repo(name=name, owner=owner)

    def org(self, name: t.Optional[str] = None):
        """
        Display the organisation menu for querying GitHub organisation data.

        :param name: Optional organisation name. If None, prompts user for input.
        """

        set_menu_title(menu_type="org")
        clear_screen()
        if name is None:
            ascii_banner(text="Org")
            name = prompts.prompt(
                message="GitHub Organisation Name",
                instruction="e.g, github",
                style=CUSTOM_STYLE,
                qmark="@",
            )

        org = Org(name=name)

        if not self.target_validator(
            identifier=name,
            instance=org,
            callback=self.org,
        ):
            return

        option = q.select(
            "What would you like to do?",
            choices=[
                q.Choice(
                    title="Change Org Name",
                    value="change_org",
                    description="Query a different organisation",
                ),
                q.Choice(
                    title="Profile",
                    value="profile",
                    description="Profile data",
                ),
                q.Choice(
                    title="Repositories",
                    value="repos",
                    description="Public repositories",
                ),
                q.Choice(
                    title="Events",
                    value="events",
                    description="Organisation events",
                ),
                q.Choice(
                    title="Hooks",
                    value="hooks",
                    description="Webhooks",
                ),
                q.Choice(
                    title="Issues",
                    value="issues",
                    description="Issues",
                ),
                q.Choice(
                    title="Members",
                    value="members",
                    description="Organisation members",
                ),
                q.Choice(
                    title="Go Back",
                    value="back",
                    description="Go back to previous menu",
                    shortcut_key="b",
                ),
                q.Choice(
                    title="Quit",
                    value="quit",
                    description="Close this session",
                    shortcut_key="q",
                ),
            ],
            style=CUSTOM_STYLE,
            instruction=INSTRUCTIONS,
            use_shortcuts=True,
        ).ask()

        if option is None:
            self.main()
            return

        # Handle navigation
        elif self.navigation_handler(option, self.org, name):
            return

        # Handle change org
        elif option == "change_org":
            self.org()
            return

        else:
            # Execute action and handle response
            self.execute_and_handle_response(
                instance=org, method_name=option, target_type="org", source=name
            )

            # After handling response, show menu again WITHOUT re-validating
            self.org(name=name)



import hashlib
import json

__all__ = ["cache"]


class ResponseCache:
    """Simple in-memory cache for API responses."""

    def __init__(self):
        """Initialise the ResponseCache with an empty cache dictionary."""

        self._cache = {}

    @staticmethod
    def _generate_key(url: str, params: dict = None) -> str:
        """
        Generate a unique cache key from URL and parameters.

        :param url: The URL to generate a key for.
        :param params: Optional dictionary of parameters to include in the key.
        :return: MD5 hash string representing the unique cache key.
        """

        cache_data = f"{url}:{json.dumps(params or {}, sort_keys=True)}"
        return hashlib.md5(cache_data.encode()).hexdigest()

    def get(self, url: str, params: dict = None):
        """
        Retrieve a cached response if it exists.

        :param url: The URL to retrieve cached data for.
        :param params: Optional dictionary of parameters used in the original request.
        :return: Cached response data if found, None otherwise.
        """

        key = self._generate_key(url, params)
        return self._cache.get(key)

    def set(self, url: str, data, params: dict = None):
        """
        Store a response in the cache.

        :param url: The URL to cache data for.
        :param data: The response data to cache.
        :param params: Optional dictionary of parameters used in the request.
        """

        key = self._generate_key(url, params)
        self._cache[key] = data

    def clear(self):
        """Clear all cached responses from memory."""
        self._cache.clear()

    def remove(self, url: str, params: dict = None):
        """
        Remove a specific cached response.

        :param url: The URL to remove cached data for.
        :param params: Optional dictionary of parameters used in the original request.
        """

        key = self._generate_key(url, params)
        self._cache.pop(key, None)


cache = ResponseCache()


def clear_cache():
    """Clear all cached API responses from the global cache instance."""

    cache.clear()



from requests import exceptions

from .github import GitHub, BASE_URL

github = GitHub()

__all__ = ["User", "Org", "Repo", "Search"]


class GitHubEntity:
    """
    Base class for GitHub entities with common functionality."""

    def __init__(self, source: str):
        """
        Initialise a GitHub entity.

        :param source: The source identifier for the entity.
        """

        self.endpoint = None
        self.source = source

    def exists(self) -> tuple[bool, dict]:
        """
        Check if the entity exists on GitHub.

        :return: Tuple of (exists, response_data) where exists is True if the entity is found.
        """
        # Check cache first
        cached = github.cache.get(self.endpoint)
        if cached is not None:
            return True, cached

        try:
            response = github.get(url=self.endpoint, return_response=True)

            if response.status_code == 200:
                data = response.json()
                # Sanitise the data
                sanitised = github.sanitise_response(data)
                # Cache the sanitised response
                github.cache.set(self.endpoint, sanitised)
                return True, sanitised

            return False, response.json()
        except exceptions.RequestException:
            return False, {}


class User(GitHubEntity):
    """Represents a GitHub user with methods to query user data."""

    def __init__(self, name: str):
        """
        Initialise a User instance.

        :param name: The GitHub username.
        """

        super().__init__(source=name)
        self.name = name
        self.endpoint = f"{BASE_URL}/users/{name}"

    def profile(self) -> dict:
        """
        Retrieve the user's profile information.

        :return: Dictionary containing user profile data.
        """

        profile = github.get(url=self.endpoint)
        return profile

    def repos(self, page: int, per_page: int) -> list:
        """
        Retrieve the user's public repositories.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of repository dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        repos = github.get(url=f"{self.endpoint}/repos", params=params)
        return repos

    def subscriptions(self, page: int, per_page: int) -> list:
        """
        Retrieve repositories the user is subscribed to.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of repository subscription dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        subscriptions = github.get(url=f"{self.endpoint}/subscriptions", params=params)
        return subscriptions

    def starred(self, page: int, per_page: int) -> list:
        """
        Retrieve repositories the user has starred.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of starred repository dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        starred = github.get(url=f"{self.endpoint}/starred", params=params)
        return starred

    def followers(self, page: int, per_page: int) -> list:
        """
        Retrieve the user's followers.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of follower user dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        users = github.get(url=f"{self.endpoint}/followers", params=params)
        return users

    def following(self, page: int, per_page: int) -> list:
        """
        Retrieve users that this user is following.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of followed user dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        users = github.get(url=f"{self.endpoint}/following", params=params)
        return users

    def follows(self, user: str) -> bool:
        """Check if this user follows another user.

        :param user: The username to check.
        :return: True if this user follows the specified user.
        """
        ...

    def orgs(self, page: int, per_page: int) -> list:
        """
        Retrieve organisations the user belongs to.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of organisation dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        orgs = github.get(url=f"{self.endpoint}/orgs", params=params)
        return orgs

    def gists(self, page: int, per_page: int) -> list:
        """
        Retrieve the user's gists.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of gist dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        gists = github.get(url=f"{self.endpoint}/gists", params=params)
        return gists

    def events(self, page: int, per_page: int) -> list:
        """
        Retrieve the user's public events.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of event dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        events = github.get(url=f"{self.endpoint}/events", params=params)
        return events

    def received_events(self, page: int, per_page: int) -> list:
        """
        Retrieve events received by the user.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of received event dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        received_events = github.get(
            url=f"{self.endpoint}/received_events", params=params
        )
        return received_events


class Org(GitHubEntity):
    """Represents a GitHub organisation with methods to query organisation data."""

    def __init__(self, name: str):
        """
        Initialise an Org instance.

        :param name: The GitHub organisation name.
        """

        super().__init__(source=name)
        self.name = name
        self.endpoint = f"{BASE_URL}/orgs/{name}"

    def profile(self) -> dict:
        """
        Retrieve the organisation's profile information.

        :return: Dictionary containing organisation profile data.
        """

        profile = github.get(url=self.endpoint)
        return profile

    def repos(self, page: int, per_page: int) -> list:
        """
        Retrieve the organisation's public repositories.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of repository dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        repos = github.get(url=f"{self.endpoint}/repos", params=params)
        return repos

    def events(self, page: int, per_page: int) -> list:
        """
        Retrieve the organisation's public events.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of event dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        events = github.get(url=f"{self.endpoint}/events", params=params)
        return events

    def hooks(self, page: int, per_page: int) -> list:
        """
        Retrieve the organisation's webhooks.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of webhook dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        hooks = github.get(url=f"{self.endpoint}/hooks", params=params)
        return hooks

    def issues(self, page: int, per_page: int) -> list:
        """
        Retrieve the organisation's issues.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of issue dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        issues = github.get(url=f"{self.endpoint}/issues", params=params)
        return issues

    def members(self, page: int, per_page: int) -> list:
        """
        Retrieve the organisation's public members.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of member user dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        members = github.get(url=f"{self.endpoint}/members", params=params)
        return members


class Repo(GitHubEntity):
    """Represents a GitHub repository with methods to query repository data."""

    def __init__(self, name: str, owner: str):
        """
        Initialise a Repo instance.

        :param name: The repository name.
        :param owner: The repository owner's username.
        """

        super().__init__(source=f"{owner}/{name}")
        self.name = name
        self.owner = owner
        self.endpoint = f"{BASE_URL}/repos/{owner}/{name}"

    def profile(self) -> dict:
        """
        Retrieve the repository's information.

        :return: Dictionary containing repository data.
        """

        profile = github.get(url=self.endpoint)
        return profile

    def forks(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's forks.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of fork repository dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        forks = github.get(url=f"{self.endpoint}/forks", params=params)
        return forks

    def issue_events(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's issue events.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of issue event dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        issue_events = github.get(url=f"{self.endpoint}/issue_events", params=params)
        return issue_events

    def events(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's events.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of event dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        events = github.get(url=f"{self.endpoint}/events", params=params)
        return events

    def assignees(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's available assignees.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of assignee user dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        assignees = github.get(url=f"{self.endpoint}/assignees", params=params)
        return assignees

    def branches(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's branches.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of branch dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        branches = github.get(url=f"{self.endpoint}/branches", params=params)
        return branches

    def tags(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's tags.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of tag dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        tags = github.get(url=f"{self.endpoint}/tags", params=params)
        return tags

    def languages(self) -> dict:
        """
        Retrieve the programming languages used in the repository.

        :return: Dictionary mapping language names to bytes of code.
        """

        languages = github.get(url=f"{self.endpoint}/languages")
        return languages

    def stargazers(self, page: int, per_page: int) -> list:
        """
        Retrieve users who have starred the repository.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of stargazer user dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        stargazers = github.get(url=f"{self.endpoint}/stargazers", params=params)
        return stargazers

    def subscribers(self, page: int, per_page: int) -> list:
        """
        Retrieve users subscribed to the repository.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of subscriber user dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        subscribers = github.get(url=f"{self.endpoint}/subscribers", params=params)
        return subscribers

    def commits(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's commits.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of commit dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        commits = github.get(url=f"{self.endpoint}/commits", params=params)
        return commits

    def comments(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's commit comments.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of comment dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        comments = github.get(url=f"{self.endpoint}/comments", params=params)
        return comments

    def contents(self, path: str) -> list:
        """
        Retrieve the contents of a file or directory in the repository.

        :param path: Path to the file or directory.
        :return: List of content item dictionaries.
        """

        contents = github.get(url=f"{self.endpoint}/contents/{path}")
        return contents

    def issues(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's issues.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of issue dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        issues = github.get(url=f"{self.endpoint}/issues", params=params)
        return issues

    def releases(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's releases.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of release dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        releases = github.get(url=f"{self.endpoint}/releases", params=params)
        return releases

    def deployments(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's deployments.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of deployment dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        deployments = github.get(url=f"{self.endpoint}/deployments", params=params)
        return deployments

    def labels(self, page: int, per_page: int) -> list:
        """
        Retrieve the repository's labels.

        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        :return: List of label dictionaries.
        """

        params = {"page": page, "per_page": per_page}
        labels = github.get(url=f"{self.endpoint}/labels", params=params)
        return labels


class Search:
    """Provides methods to search GitHub for various entity types."""

    def __init__(self, query: str, page: int, per_page: int):
        """
        Initialise a Search instance.

        :param query: The search query string.
        :param page: Page number for pagination.
        :param per_page: Number of results per page.
        """

        self.query = query
        self.page = page
        self.per_page = per_page
        self.endpoint = f"{BASE_URL}/search"
        self.params = {"q": query, "page": page, "per_page": per_page}

    def repos(self) -> list:
        """
        Search for repositories matching the query.

        :return: List of repository search result dictionaries.
        """

        repos = github.get(url=f"{self.endpoint}/repositories", params=self.params)
        return repos

    def users(self) -> list:
        """
        Search for users matching the query.

        :return: List of user search result dictionaries.
        """

        users = github.get(url=f"{self.endpoint}/users", params=self.params)
        return users

    def commits(self) -> list:
        """
        Search for commits matching the query.

        :return: List of commit search result dictionaries.
        """

        commits = github.get(url=f"{self.endpoint}/commits", params=self.params)
        return commits

    def issues(self) -> list:
        """
        Search for issues and pull requests matching the query.

        :return: List of issue search result dictionaries.
        """

        issues = github.get(url=f"{self.endpoint}/issues", params=self.params)
        return issues

    def topics(self) -> list:
        """
        Search for topics matching the query.

        :return: List of topic search result dictionaries.
        """

        topics = github.get(url=f"{self.endpoint}/topics", params=self.params)
        return topics



import typing as t

import questionary as q
from questionary import Style

__all__ = ["Prompts"]


class Prompts:
    """Provides interactive prompts for user input collection."""

    def __init__(self):
        """Initialise the Prompts class."""
        pass

    @staticmethod
    def prompt(
        message: str,
        instruction: t.Optional[str] = None,
        style: t.Optional[Style] = None,
        qmark: t.Optional[str] = "?",
    ) -> str:
        """
        Display a text input prompt and validate non-empty input.

        :param message: The prompt message to display.
        :param instruction: Optional instruction text to guide the user.
        :param style: Optional questionary Style object for custom styling.
        :param qmark: Optional custom question mark character or string.
        :return: The user's input string.
        :raises KeyboardInterrupt: If the user cancels the prompt with CTRL+C.
        """

        result = q.text(
            message=message,
            instruction=instruction,
            style=style,
            qmark=qmark,
            validate=lambda text: len(text.strip()) > 0
            or None
            or "Input cannot be empty",
        ).ask()

        if result is None:
            raise KeyboardInterrupt

        return result

    @staticmethod
    def pagination_params() -> dict:
        """
        Prompt the user for pagination parameters.

        :return: Dictionary containing 'page' and 'per_page' integer values.
        """

        try:
            page = q.text(message="Page", default="1", qmark="n").ask()
            per_page = q.text(
                message="Per Page",
                default="100",
                qmark="n",
            ).ask()

            return {
                "page": int(page) if page else 1,
                "per_page": min(int(per_page) if per_page else 100, 100),
            }

        except (ValueError, TypeError):
            print("Invalid input, using defaults (page=1, per_page=100)")
            return {"page": 1, "per_page": 100}
